# -*- coding: utf-8 -*-
"""
This module implements the CLI tool 'niddel' that can be used to interact with the Niddel Magnet
v2 API using magnetsdk2.
"""

import argparse
import json
import logging
from codecs import BOM_UTF8
from datetime import datetime
from errno import EPIPE
from glob import glob
from os import linesep, sep
from os.path import expanduser, join, basename, isfile
from sys import stdout, stderr, exc_info
from uuid import UUID

import boto3
import six

from magnetsdk2 import Connection, __version__
from magnetsdk2.cef import convert_alert
from magnetsdk2.iterator import FilePersistentAlertIterator
from magnetsdk2.time import UTC
from magnetsdk2.validation import parse_date

# logging setup
logger = logging.getLogger('magnetsdk2')
handler = logging.StreamHandler(stderr)
handler.setFormatter(
    logging.Formatter('%(asctime)s pid=%(process)d %(module)s %(levelname)s %(message)s',
                      '%Y-%m-%dT%H:%M:%S%z'))
logger.addHandler(handler)


def main():
    # top-level parser
    parser = argparse.ArgumentParser(prog='niddel',
                                     description='Command-line utility to interact with the ' +
                                                 'Niddel Magnet v2 API (v{0:s})'.format(__version__))
    parser.add_argument("-p", "--profile",
                        help="which profile (from ~/.magnetsdk/config) to obtain API key from",
                        default='default')
    parser.add_argument("-i", "--indent", help="indent JSON output", action="store_const", const=4)
    parser.add_argument("-v", "--verbose", help="set verbose mode", action="store_true",
                        default=False)
    parser.add_argument("-o", "--outfile",
                        help="destination file to write to, if exists will be overwritten",
                        type=argparse.FileType('wb'), default=stdout)
    parser.set_defaults(indent=None, parser=parser)
    subparsers = parser.add_subparsers()

    # "me" command
    me_parser = subparsers.add_parser('me', help="display API key owner information",
                                      description="display API key owner information")
    me_parser.set_defaults(func=command_me, parser=me_parser)

    # "organizations" command
    org_parser = subparsers.add_parser('organizations',
                                       help="list basic organization information",
                                       description="list basic organization information")
    org_parser.add_argument("--id", help="get details on organization with the provided ID",
                            type=UUID, required=False)
    org_parser.set_defaults(func=command_organizations, id=None, parser=org_parser)

    # "alerts" command
    alerts_parser = subparsers.add_parser('alerts',
                                          help="list an organization's alerts",
                                          description="list an organization's alerts")
    alerts_parser.add_argument("organization", help="ID of the organization",
                               type=UUID)
    alerts_parser.add_argument("--start", help="initial batch date to process in YYYY-MM-DD format",
                               type=parse_arg_date)
    alerts_parser.add_argument("-p", "--persist",
                               help="file to store persistent state data, to ensure only alerts " +
                                    "that haven't been seen before are part of the output")
    alerts_parser.add_argument("-f", "--format", choices=['json', 'cef'], default='json',
                               help="format in which to output alerts")
    alerts_parser.set_defaults(func=command_alerts, start=None, persist=None, parser=alerts_parser)

    # "logs" command
    logs_parser = subparsers.add_parser('logs',
                                        help="upload, download or list log files",
                                        description="use temporary credentials to access " +
                                                    "log files to an organization's assigned " +
                                                    "S3 bucket's upload folder")
    logs_parser.add_argument("organization", help="ID of the organization", type=UUID)
    logs_parser.set_defaults(parser=logs_parser)

    logs_subparsers = logs_parser.add_subparsers()

    # "logs list" command
    logs_list_parser = logs_subparsers.add_parser('list', help='list log files',
                                                  description='list files in the organization\'s ' +
                                                              'upload folder')
    logs_list_parser.add_argument("-f", "--format", choices=['json', 'table'], default='table',
                                  help="format in which to output alerts")

    logs_list_parser.set_defaults(func=command_logs_list, parser=logs_list_parser)

    # "logs upload" command
    logs_upload_parser = logs_subparsers.add_parser('upload', help='upload log files',
                                                    description='upload log files to the ' +
                                                                'organization\'s upload folder')
    logs_upload_parser.add_argument("-f", "--folder",
                                    help="sub-folder of the upload folder to send file to")
    logs_upload_parser.add_argument("-p", "--prefix", choices=['day', 'hour'],
                                    required=False,
                                    help="prefix destination file name with UTC date in " +
                                         "YYYY-MM-DD format or hour in YYYY-MM-DD-HH format")
    logs_upload_parser.add_argument("src", help="source file name(s) or wildcard(s)", nargs="+")
    logs_upload_parser.set_defaults(func=command_logs_upload, parser=logs_upload_parser)

    # parse arguments
    args = parser.parse_args()
    if args.verbose:
        logger.setLevel(logging.DEBUG)

    # open connection and dispatch to proper function
    try:
        conn = Connection(profile=args.profile)
        args.func(conn, args)
    except Exception as e:
        logger.debug("exception caught in processing", exc_info=True)
        args.parser.error(e.message)
    else:
        if args.outfile != stdout:
            args.outfile.close()


def parse_arg_date(value):
    try:
        return parse_date(value)
    except Exception as e:
        logger.debug(e)
        raise argparse.ArgumentTypeError("unable to parse date, YYYY-MM-DD format expected")


def parse_glob_files(x):
    retval = set()
    if isinstance(x, six.string_types):
        retval |= set(glob(expanduser(x)))
    elif isinstance(x, list):
        for subx in x:
            retval |= parse_glob_files(subx)
    else:
        raise ValueError('unexpected type')
    return retval


def command_me(conn, args):
    json.dump(conn.get_me(), args.outfile, indent=args.indent)


def command_organizations(conn, args):
    if args.id:
        json.dump(conn.get_organization(args.id.__str__()), args.outfile, indent=args.indent)
    else:
        for organization in conn.iter_organizations():
            try:
                json.dump(organization, args.outfile, indent=args.indent)
                args.outfile.write(linesep)
            except IOError as ioe:
                if ioe.errno == EPIPE and args.outfile == stdout:
                    logger.debug('stdout closed, exiting...')
                    break
                else:
                    six.reraise(*exc_info())


def command_alerts(conn, args):
    if args.persist:
        iterator = FilePersistentAlertIterator(filename=args.persist, connection=conn,
                                               organization_id=args.organization.__str__(),
                                               start_date=args.start)
    else:
        iterator = conn.iter_organization_alerts(organization_id=args.organization.__str__(),
                                                 fromDate=args.start, sortBy='batchDate')

    if args.outfile != stdout and args.format == 'cef':
        args.outfile.write(BOM_UTF8)

    for alert in iterator:
        try:
            if args.format == 'json':
                json.dump(alert, args.outfile, indent=args.indent)
            elif args.format == 'cef':
                convert_alert(args.outfile, alert, args.organization.__str__())
            args.outfile.write(linesep)
        except IOError as ioe:
            if ioe.errno == EPIPE and args.outfile == stdout:
                logger.debug('stdout closed, exiting...')
                break
            else:
                six.reraise(*exc_info())

    if args.persist:
        iterator.save()


def command_logs_list(conn, args):
    # connect to S3
    creds = conn.get_organization_credentials(args.organization.__str__())
    bucket = boto3.resource('s3', aws_access_key_id=creds['accessKeyId'],
                            aws_secret_access_key=creds['secretAccessKey'],
                            aws_session_token=creds['sessionToken'],
                            region_name=creds['bucketRegion']).Bucket(creds['bucket'])
    logger.debug('opened S3 bucket %s in %s successfully', creds['bucket'], creds['bucketRegion'])
    prefix = conn.get_organization(args.organization.__str__())['properties']['bucketUploadPrefix']

    # list objects
    for key in bucket.objects.filter(Prefix=prefix + '/'):
        name = 's3://{0:s}/{1:s}'.format(key.bucket_name, key.key)
        last_modified = key.last_modified.strftime('%c')
        try:
            if args.format == 'json':
                json.dump({'name': name, 'size': key.size, 'last_modified': last_modified},
                          args.outfile, indent=args.indent)
            elif args.format == 'table':
                args.outfile.write('{0:s} {1:-12d} {2:s}'.format(last_modified, key.size, name))
            args.outfile.write(linesep)
        except IOError as ioe:
            if ioe.errno == EPIPE and args.outfile == stdout:
                logger.debug('stdout closed, exiting...')
                break
            else:
                six.reraise(*exc_info())


def command_logs_upload(conn, args):
    # get all source files and perform basic sanity checking
    srcfiles = parse_glob_files(args.src)
    if not srcfiles:
        raise Exception('no valid source files found')
    notfiles = {'"' + x + '"' for x in srcfiles if not isfile(x)}
    if notfiles:
        raise Exception('invalid files found: {0:s}'.format(', '.join(notfiles)))
    del notfiles

    # connect to S3
    creds = conn.get_organization_credentials(args.organization.__str__())
    bucket = boto3.resource('s3', aws_access_key_id=creds['accessKeyId'],
                            aws_secret_access_key=creds['secretAccessKey'],
                            aws_session_token=creds['sessionToken'],
                            region_name=creds['bucketRegion']).Bucket(creds['bucket'])
    logger.debug('opened S3 bucket %s in %s successfully', creds['bucket'], creds['bucketRegion'])
    uploadprefix = conn.get_organization(args.organization.__str__())['properties'][
        'bucketUploadPrefix']

    # determine filename prefix to use, if any
    if args.prefix == 'day':
        slotprefix = datetime.now(UTC).strftime("%Y-%m-%d_")
    elif args.prefix == 'hour':
        slotprefix = datetime.now(UTC).strftime("%Y-%m-%d-%H_")
    else:
        slotprefix = ''

    # process each file
    if sep != '/':
        uploadprefix = uploadprefix.replace('/', sep)
    for src in srcfiles:
        dest = join(uploadprefix, slotprefix + basename(src))
        if sep != '/':
            dest = dest.replace(sep, '/')

        try:
            args.outfile.write(
                'copying {0:s} to s3://{1:s}/{2:s} ...'.format(src, creds['bucket'], dest))
        except IOError as ioe:
            if ioe.errno == EPIPE and args.outfile == stdout:
                logger.debug('stdout closed, exiting...')
                break
            else:
                six.reraise(*exc_info())

        bucket.upload_file(src, dest, ExtraArgs={'ServerSideEncryption': 'AES256'})

        try:
            args.outfile.write(' Done.' + linesep)
        except IOError as ioe:
            if ioe.errno == EPIPE and args.outfile == stdout:
                logger.debug('stdout closed, exiting...')
                break
            else:
                six.reraise(*exc_info())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
This module allows persistent iteration of alerts. Some use cases include opening tickets based
on new alerts, or even automating responses for some high-confidence alerts.
"""
import json
from abc import ABCMeta, abstractmethod
from collections import Iterable, Iterator
from os.path import isfile

from six import python_2_unicode_compatible

from magnetsdk2.connection import Connection
from magnetsdk2.validation import is_valid_uuid, parse_date


@python_2_unicode_compatible
class PersistenceEntry(object):
    """Class that encapsulates the minimal persistence information needed to continously process
    only new alerts from a given organization."""

    @property
    def organization_id(self):
        """A string in UUID format containing the ID of the organization."""
        return self._organization_id

    @property
    def latest_batch_date(self):
        """A string representing a date in ISO8601 format, which indicates the most recent batch
        date that has already been processed."""
        return self._latest_batch_date

    @latest_batch_date.setter
    def latest_batch_date(self, latest_batch_date):
        if latest_batch_date is None:
            self._latest_batch_date = None
        else:
            self._latest_batch_date = parse_date(latest_batch_date)

    @property
    def latest_alert_ids(self):
        """A set of string in UUID format that indicate which alerts on the latest batch date
        have already been processed."""
        return self._latest_alert_ids

    @latest_alert_ids.setter
    def latest_alert_ids(self, latest_alert_ids):
        if latest_alert_ids is None:
            self._latest_alert_ids = set()
        else:
            if not isinstance(latest_alert_ids, Iterable):
                raise ValueError('latest alert IDs must be iterable')
            if latest_alert_ids and not all(is_valid_uuid(x) for x in latest_alert_ids):
                raise ValueError('latest alert IDs must only contain UUIDs')
            self._latest_alert_ids = {x for x in latest_alert_ids}

    def add_alert_id(self, alert_id):
        if not is_valid_uuid(alert_id):
            raise ValueError("invalid alert ID")
        self._latest_alert_ids.add(alert_id)

    def __init__(self, organization_id, latest_batch_date=None, latest_alert_ids=None):
        if not is_valid_uuid(organization_id):
            raise ValueError('invalid organization ID')
        self._organization_id = organization_id

        self._latest_alert_ids = set()
        self._latest_batch_date = None
        self.latest_batch_date = latest_batch_date
        self.latest_alert_ids = latest_alert_ids

    def __str__(self):
        return "%s(organization_id=%s, latest_batch_date=%s, latest_alert_ids=%s)" \
               % (self.__class__.__name__, self.organization_id, self.latest_batch_date,
                  self.latest_alert_ids)


@python_2_unicode_compatible
class AbstractPersistentAlertIterator(Iterator):
    """Abstract class that encapsulates the logic of walking through an organization's alerts in
    increasing batch date order and persisting state the prevents an alert from being seen multiple
    times across executions."""

    __metaclass__ = ABCMeta

    def __init__(self, connection, organization_id, start_date=None):
        """Initializes a persistent alert iterator.
        :param connection: an instance of magnetsdk2.Connection
        :param organization_id: a string containing an organization ID in UUID format
        :param start_date: optional date that represents the initial batch date to load alerts from
        """
        if not isinstance(connection, Connection):
            raise ValueError('invalid connection')
        self._connection = connection

        if not is_valid_uuid(organization_id):
            raise ValueError('invalid organization ID')
        self._organization_id = organization_id

        if start_date is not None:
            self._start_date = parse_date(start_date)
        else:
            self._start_date = None
        self._persistence_entry = None
        self._alerts = []

    @property
    def organization_id(self):
        return self._organization_id

    @property
    def start_date(self):
        return self._start_date

    @property
    def connection(self):
        return self._connection

    @property
    def persistence_entry(self):
        if not self._persistence_entry:
            self._persistence_entry = self._load()
            if self._persistence_entry is None:
                self._persistence_entry = PersistenceEntry(self.organization_id)
            else:
                if not isinstance(self._persistence_entry, PersistenceEntry):
                    raise ValueError('load method should return a PersistenceEntry instance')
                if not self._persistence_entry.organization_id == self.organization_id:
                    raise ValueError(
                        'PersistenceEntry instance does not match organization ID ' \
                        + self.organization_id)

            if self._start_date:
                if self._persistence_entry.latest_batch_date is None \
                        or self._persistence_entry.latest_batch_date < self._start_date:
                    self._persistence_entry.latest_batch_date = self._start_date
                    self._persistence_entry.latest_alert_ids = None

        return self._persistence_entry

    @abstractmethod
    def _load(self):
        """Abstract method that loads a given organization's persistence data. Implementations
        should return a PersistenceEntry instance or None."""
        pass

    @abstractmethod
    def _save(self):
        """Abstract method that saves a given organization's persistence data."""
        pass

    def __iter__(self):
        return self

    def _load_alerts(self):
        # if we already have cached alerts, do nothing
        if self._alerts:
            return

        # get the candidate dates in order, and discard the ones we've already fully processed
        dates = sorted([x for x in
                        self.connection.list_organization_alert_dates(
                            self.persistence_entry.organization_id,
                            'batchDate')])
        if self.persistence_entry.latest_batch_date:
            dates = [x for x in dates if x >= self.persistence_entry.latest_batch_date]

        # loop over candidate dates
        for d in dates:
            # if candidate date is newer, reset persistence data to it
            if d != self._persistence_entry.latest_batch_date:
                self._persistence_entry.latest_batch_date = d
                self._persistence_entry.latest_alert_ids = None

            # add any alerts on the candidate date we haven't processed yet to the cache
            for alert in self._connection.iter_organization_alerts(
                    organization_id=self._persistence_entry.organization_id,
                    fromDate=d, toDate=d, sortBy='batchDate'):
                if alert['id'] not in self._persistence_entry.latest_alert_ids:
                    self._alerts.append(alert)

            # if alert cache is not empty, we are finished for now
            if self._alerts:
                return

    def save(self):
        self._save()

    def load(self):
        self._persistence_entry = None
        self._alerts = []

    def next(self):
        if not self._alerts:
            self._load_alerts()

        if self._alerts:
            alert = self._alerts.pop()
            self._persistence_entry.latest_batch_date = alert['batchDate']
            self._persistence_entry.add_alert_id(alert['id'])
            return alert
        else:
            raise StopIteration

    def __str__(self):
        return "%s(organization_id=%s, start_date=%s, persistence_entry=%s)" \
               % (self.__class__.__name__, self.organization_id, self.start_date,
                  self.persistence_entry)


@python_2_unicode_compatible
class FilePersistentAlertIterator(AbstractPersistentAlertIterator):
    """Subclass of AbstractPersistentAlertIterator that saves the persistence state as a JSON object
    on a given text file. Does not lock the file or otherwise prevent multiple processes corrupting
    the state, so any such control is left up to the caller. Assumes one file per organization, as
    the save method completely overwrites the file contents."""

    def __init__(self, filename, *args, **kwargs):
        self._filename = filename
        super(FilePersistentAlertIterator, self).__init__(*args, **kwargs)

    @property
    def filename(self):
        return self._filename

    def _load(self):
        if isfile(self._filename):
            with open(self._filename, 'r') as f:
                pe = json.load(f)
                return PersistenceEntry(pe['organization_id'], pe['latest_batch_date'],
                                        pe['latest_alert_ids'])
        else:
            return None

    def _save(self):
        pe = {
            'organization_id': self.persistence_entry.organization_id,
            'latest_batch_date': self.persistence_entry.latest_batch_date,
            'latest_alert_ids': [x for x in self.persistence_entry.latest_alert_ids]
        }
        with open(self._filename, 'w') as f:
            json.dump(pe, f)

    def __str__(self):
        return super(FilePersistentAlertIterator, self).__str__()[:-1] + ", filename=%s)" \
                                                                         % self._filename



# -*- coding: utf-8 -*-
"""
This module implements basic validation and conversion logic for API data.
"""
import datetime
import re
from collections import Iterable

import iso8601
import rfc3987
import six

_UUID_REGEX = re.compile(
    "(?i)^[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$")


def is_valid_uuid(value):
    """Validates if a value is a string representation of a UUID.
    :param value: string to validate
    :return: a boolean
    """
    return isinstance(value, six.string_types) and _UUID_REGEX.match(value)


def is_valid_uri(value):
    """Validates if a value is a valid string with an URI.
    :param value: string to validate
    :return: a boolean
    """
    return isinstance(value, six.string_types) and rfc3987.match(value, 'URI')


def is_valid_port(value):
    """Validates if a value is a valid integer with a TCP/UDP port number.
    :param value: string to validate
    :return: a boolean
    """
    return isinstance(value, six.integer_types) and 1 <= value <= 65535


def is_valid_alert_sortBy(value):
    """Validates if a value is a valid string with an alert sortBy parameter value as per the
    Magnet API v2 specification.
    :param value: string to validate
    :return: a boolean
    """
    return isinstance(value, six.string_types) and value in ('logDate', 'batchDate')


def is_valid_alert_status(value):
    """Validates if a value is a valid string with an alert status value as per the Magnet API v2
    specification.
    :param value: string to validate
    :return: a boolean
    """
    if not isinstance(value, Iterable):
        return False
    for string in value:
        if not isinstance(string, six.string_types) or string not in ('new', 'under_investigation',
                                                                      'rejected', 'resolved'):
            return False
    return True


def parse_date(value):
    """Extracts a datetime.date object from a string containing an ISO 8601 date, a
    datetime.datetime object or a datetime.date object.
    :param value: string to parse, datetime.date or datetime.datetime instance
    :return: a datetime.date instance
    """
    if isinstance(value, six.string_types):
        return iso8601.parse_date(value).date().isoformat()
    elif isinstance(value, datetime.datetime):
        return value.date().isoformat()
    elif isinstance(value, datetime.date):
        return value.isoformat()
    else:
        raise ValueError('date must be in ISO format: ' + repr(value))



# -*- coding: utf-8 -*-
"""
This module implements the Connection class, which is used for low-level interaction with the
Niddel Magnet v2 API.
"""
import logging
import os
import sys

import six
from requests import request
from six.moves.configparser import RawConfigParser
from six.moves.urllib.parse import urlsplit, quote_plus

from magnetsdk2.time import UTC
from magnetsdk2.validation import is_valid_uuid, is_valid_uri, is_valid_port, \
    is_valid_alert_sortBy, is_valid_alert_status, parse_date

# Default values used for the configuration
_CONFIG_DIR = os.path.expanduser('~/.magnetsdk')
_CONFIG_FILE = os.path.join(_CONFIG_DIR, "config")
_DEFAULT_CONFIG = {'endpoint': 'https://api.niddel.com/v2'}
_API_KEY_HEADER = 'X-Api-Key'
_PAGE_SIZE = 100


class Connection(object):
    """ This class encapsulates accessing the Niddel Magnet v2 API (https://api.niddel.com/v2)
     using a particular configuration profile from ~/.magnetsdk/config, and is wrapper around
     the requests library that is used for all accesses.
    """

    def __init__(self, profile='default', api_key=None, endpoint=None):
        """ Initializes the connection with the proper configuration data.
        :param profile: the profile name to use in ~/.magnetsdk/config
        :param api_key: if provided, this API key is used instead of the one on the
        configuration file
        :param endpoint: if provided this endpoint URL is used instead of the one on the
        configuration file
        """
        # initialize logger and credential cache
        self._logger = logging.getLogger('magnetsdk2')
        self._org_creds_cache = {}

        # initially get configuration from environment
        self.endpoint = os.getenv('MAGNETSDK_API_ENDPOINT', _DEFAULT_CONFIG['endpoint'])
        self.api_key = os.getenv('MAGNETSDK_API_KEY')

        # read from configuration file and profile
        if profile and os.path.isfile(_CONFIG_FILE):
            parser = RawConfigParser(_DEFAULT_CONFIG)
            parser.read(_CONFIG_FILE)
            if not parser.has_section(profile):
                raise ValueError('profile %s not found in %s' % (profile, _CONFIG_FILE))
            self.api_key = parser.get(profile, 'api_key')
            self.endpoint = parser.get(profile, 'endpoint')

        # explicit parameters override whatever was read previously
        if api_key is not None:
            if not isinstance(api_key, six.string_types):
                raise ValueError("API key must be a string")
            self.api_key = api_key
        if endpoint is not None:
            if not is_valid_uri(endpoint):
                raise ValueError("endpoint must be a string containing a valid URL")
            self.endpoint = endpoint

        # ensure we have an API key to work with
        if not self.api_key:
            raise ValueError('no API key to use')

        # parse endpoint data and ensure it ends with a slash
        endpoint = urlsplit(self.endpoint, 'https')
        self.endpoint = endpoint.geturl()
        if not self.endpoint.endswith('/'):
            self.endpoint += '/'

        # check for certificate pinning for the destination server
        if os.path.isfile(os.path.join(_CONFIG_DIR, endpoint.hostname + '.pem')):
            self.verify = os.path.join(_CONFIG_DIR, endpoint.hostname + '.pem')
        else:
            self.verify = True

        self._logger.debug('%s: endpoint=%r, verify=%r', self.__class__.__name__, self.endpoint,
                           self.verify)
        self._proxies = None

    def __del__(self):
        self.close()

    def set_proxy(self, proxy, proxy_port=None, proxy_user=None, proxy_pass=None,
                  proxy_proto='https'):
        """Configure this connection to use an HTTPS proxy to access the API endpoint.
        :param proxy: string containing the proxy hostname or iP address
        :param proxy_port: integer containing the proxy port to connect to (optional)
        :param proxy_user: string containing the username to use for basic authentication to the
        proxy (optional)
        :param proxy_pass: string containing the password to use for basic authentication to the
        proxy (optional)
        :param proxy_proto: string containing the proxy protocol equal to either 'http' or 'https'
        :return: the proxy URL
        """
        if not isinstance(proxy_proto, six.string_types) or proxy_proto not in ('http', 'https'):
            raise ValueError("proxy protocol must be a string")
        proxy_url = proxy_proto + '://'
        proxy_url_sanitized = proxy_proto + '://'
        if proxy_user and proxy_pass:
            if not isinstance(proxy_user, six.string_types):
                raise ValueError("proxy username must be a string")
            if not isinstance(proxy_pass, six.string_types):
                raise ValueError("proxy password must be a string")
            proxy_url += quote_plus(proxy_user) + ':' + quote_plus(proxy_pass) + '@'
            proxy_url_sanitized += quote_plus(proxy_user) + ':' + ('*' * len(proxy_pass)) + '@'
        elif proxy_user:
            if not isinstance(proxy_user, six.string_types):
                raise ValueError("proxy username must be a string")
            proxy_url += quote_plus(proxy_user) + '@'
            proxy_url_sanitized += quote_plus(proxy_user) + '@'
        if not isinstance(proxy, six.string_types):
            raise ValueError("proxy hostname or IP address must be a string")
        proxy_url += proxy
        proxy_url_sanitized += proxy
        if proxy_port:
            if isinstance(proxy_port, six.string_types):
                proxy_port = int(proxy_port)
            if not is_valid_port(proxy_port):
                raise ValueError("invalid proxy port")
            proxy_url += ":%i" % proxy_port
            proxy_url_sanitized += ":%i" % proxy_port
        self.set_proxy_url(proxy_url)
        self._logger.debug('using proxy URL ' + proxy_url_sanitized)
        return proxy_url

    def set_proxy_url(self, proxy_url):
        """Configure this connection to use an HTTPS proxy to access the API endpoint using a
        URL as per http://docs.python-requests.org/en/master/user/advanced/#proxies
        :param proxy_url: string containing the proxy URL
        """
        self._proxies = {
            'http': proxy_url,
            'https': proxy_url
        }

    def clear_proxy(self):
        """Removes the existing proxy configuration so that """
        if self._proxies:
            self.close()
            self._proxies = None

    def close(self):
        """ Closes the Connection object.
        """
        pass

    def _request(self, method, path, params=None, body=None):
        """ Performs an HTTP operation using the base API endpoint, API key and SSL validation /
        cert pinning obtained from the configuration file.
        :param method: string with the the HTTP method to use ('GET', 'PUT', etc.)
        :param path: string with the path to append to the base API endpoint
        :param params: dict with the query parameters to submit
        :return: the requests.Response object
        """
        response = request(method=method, url=self.endpoint + path, params=params, json=body,
                           verify=self.verify,
                           proxies=self._proxies, timeout=(5, 60),
                           headers={_API_KEY_HEADER: self.api_key,
                                    "Accept-Encoding": "gzip, deflate",
                                    "User-Agent": "magnet-sdk-python",
                                    "Accept": "application/json"})
        if response.request.body:
            msg = '{0:s} {1:s} ({2:d} bytes in body)'.format(response.request.method,
                                                             response.request.url,
                                                             len(response.request.body))
            self._logger.debug(msg)
        else:
            self._logger.debug('{0:s} {1:s}'.format(response.request.method, response.request.url))
        self._logger.debug("got {0:d} response".format(response.status_code))
        return response

    def _request_retry(self, method, path, params=None, body=None, ok_status=(200, 404), retries=5):
        """ Wrapper around self._request that retries on exceptions and unexpected status codes.
        """
        i = 1
        while True:
            try:
                response = self._request(method, path, params, body)
                if i >= retries or response.status_code in ok_status:
                    return response
            except:
                self._logger.exception('error at try %i of %s request for %s with params=%s', i,
                                       method, path, repr(params))
                if i >= retries:
                    six.reraise(*sys.exc_info())
            i += 1

    def iter_organizations(self):
        """ Generator that allows iteration over all of the organizations that this connections's
        API key has access to.
        :return: an iterator over the decoded JSON objects that represent organizations.
        """
        params = {
            'page': 1,
            'size': _PAGE_SIZE
        }
        while True:
            response = self._request_retry("GET", path='organizations', params=params)
            if response.status_code == 200:
                organization_list = response.json()
                for organization in organization_list:
                    yield organization
                if len(organization_list) < _PAGE_SIZE:
                    return
            elif response.status_code == 404:
                return
            else:
                response.raise_for_status()
            params['page'] += 1

    def get_organization(self, organization_id):
        """ Retrieves detailed data from an organization this API key has accessed to based on its
        ID.
        :param organization_id: string with the UUID-style unique ID of the organization
        :return: decoded JSON objects that represents the organization or None if not found
        """
        if not is_valid_uuid(organization_id):
            raise ValueError("organization id should be a string in UUID format")
        response = self._request_retry("GET", path='organizations/' + organization_id)
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 404:
            return None
        else:
            response.raise_for_status()

    def get_organization_credentials(self, organization_id, cache=True):
        """ Retrieves a new set of temporary AWS credentials to allow access to an organization's
        S3 bucket. Typically used to upload log files. Will cache the response and return the same
        credentials if they have at least 10 minutes before expiration.
        :param organization_id: string with the UUID-style unique ID of the organization
        :param cache: boolean controlling whether credentials are cached in this connection
        :return: decoded JSON objects that represents the credentials
        """
        if not is_valid_uuid(organization_id):
            raise ValueError("organization id should be a string in UUID format")

        # try cached response first, must have at least 10 minutes validity
        if cache:
            cached_response = self._org_creds_cache.get(organization_id, None)
            if cached_response:
                exp = iso8601.parse_date(cached_response['expiration'])
                if exp >= (datetime.datetime.now(UTC) + datetime.timedelta(minutes=10)):
                    return cached_response
                else:
                    del self._org_creds_cache[organization_id]

        # get a new credential
        response = self._request_retry("GET", path='organizations/%s/credentials' % organization_id)
        if response.status_code == 200:
            cached_response = response.json()
            if cache:
                self._org_creds_cache[organization_id] = cached_response
            return cached_response
        else:
            response.raise_for_status()

    def iter_organization_alerts(self, organization_id, fromDate=None, toDate=None,
                                 sortBy="logDate", status=None):
        """ Generator that allows iteration over an organization's alerts, with optional filters.
        :param organization_id: string with the UUID-style unique ID of the organization
        :param fromDate: only list alerts with dates >= this parameter
        :param toDate: only list alerts with dates <= this parameter
        :param sortBy: one of 'logDate' or 'batchDate', controls which date field fromDate and
        toDate apply to
        :param status: a list or set containing one or more of 'new', 'under_investigation',
        'rejected', 'resolved'
        :return: an iterator over the decoded JSON objects that represent alerts.
        """
        if not is_valid_uuid(organization_id):
            raise ValueError("organization id should be a string in UUID format")
        if not is_valid_alert_sortBy(sortBy):
            raise ValueError("sortBy must be either 'logDate' or 'batchDate'")
        if status is not None and not is_valid_alert_status(status):
            raise ValueError(
                "status must be an iterable with one or more of 'new', 'under_investigation', " +
                "'rejected' or 'resolved'")

        # loop over alert pages and yield them
        params = {
            'page': 1,
            'size': _PAGE_SIZE,
            'sortBy': sortBy
        }
        if fromDate:
            params['fromDate'] = parse_date(fromDate)
        if toDate:
            params['toDate'] = parse_date(toDate)
        if status:
            params['status'] = status

        while True:
            response = self._request_retry("GET", path='organizations/%s/alerts' % organization_id,
                                           params=params)
            if response.status_code == 200:
                alert_list = response.json()
                for alert in alert_list:
                    yield alert
                if len(alert_list) < _PAGE_SIZE:
                    return
            elif response.status_code == 404:
                return
            else:
                response.raise_for_status()
            params['page'] += 1

    def list_organization_alert_dates(self, organization_id, sortBy="logDate"):
        """ Lists all log or batch dates for which alerts exist on the organization.
        :param organization_id: string with the UUID-style unique ID of the organization
        :param sortBy: one of 'logDate' or 'batchDate', controls which date field to return
        :return: a set of ISO 8601 dates for which alerts exist
        """
        if not is_valid_uuid(organization_id):
            raise ValueError("organization id should be a string in UUID format")
        if not is_valid_alert_sortBy(sortBy):
            raise ValueError("sortBy must be either 'logDate' or 'batchDate'")

        response = self._request_retry("GET",
                                       path='organizations/%s/alerts/dates' % organization_id,
                                       params={'sortBy': sortBy})
        if response.status_code == 200:
            return set(response.json())
        elif response.status_code == 404:
            return set()
        else:
            response.raise_for_status()

    def get_me(self):
        """Queries the API about the user that owns the API key in use.
        :return: a dict representing the user details
        """
        response = self._request_retry("GET", path="me", ok_status=(200,))
        if response.status_code == 200:
            return response.json()
        else:
            response.raise_for_status()



# -*- coding: utf-8 -*-
"""
This module implements writing CEF format events.
"""

from math import ceil, trunc

import six

from magnetsdk2.time import seconds_from_UTC_epoch


def escape_header_entry(x):
    """
    Escapes backslashes and pipes from a header entry.
    :param x: the string value to escape
    :return: escaped and trimmed UTF-8 encoded str / bytes
    """
    if not isinstance(x, six.string_types):
        x = x.__str__()
    return x.replace('\\', '\\\\').replace('|', '\\|').strip()


def header(device_vendor, device_product, device_version, signature_id, name, severity):
    """
    Builds a CEF version 0 header with the given fields
    :return: escaped and trimmed UTF-8 encoded str / bytes
    """
    if isinstance(severity, float):
        severity = trunc(severity)
    if isinstance(severity, six.integer_types):
        if severity < 0 or severity > 10:
            raise ValueError('severity must be between 0 and 10')
        severity = '{0:d}'.format(severity)
    return '|'.join(map(escape_header_entry,
                        ['CEF:0', device_vendor, device_product, device_version, signature_id, name,
                         severity, ''])).strip()


def escape_extension_value(x):
    """
    Escapes backslashes, pipes, equals signs and newlines from an extension entry value.
    :param x: the string value to escape
    :return: escaped and trimmed UTF-8 encoded str / bytes
    """
    if not isinstance(x, six.string_types):
        x = x.__str__()
    return x.replace('\\', '\\\\').replace('=', '\\=').replace('\n', '\\n').replace('\r',
                                                                                    '\\r').strip()


def extension(fields):
    """
    Builds a CEF version 0 extension with the given fields. Fields will be sorted by name.
    :param fields: dict containing fields to include
    :return: escaped and trimmed UTF-8 encoded str / bytes
    """
    fields = sorted([(k, v) for k, v in six.iteritems(fields) if v], key=lambda x: x[0])
    return ' '.join([e[0].strip() + '=' + escape_extension_value(e[1])
                     for e in fields]).strip()


def timestamp(ts):
    """
    Converts an ISO date and time in UTC into milliseconds from epoch as expected by CEF format.
    :param ts: string containing the date and time in ISO 8601 format
    :return: number of milliseconds since epoch
    """
    if not ts:
        return None
    if not ts.endswith('Z'):
        ts = ts + 'Z'
    return '{0:d}'.format(trunc(seconds_from_UTC_epoch(ts) * 1000))


def convert_alert(obj, alert, organization):
    """
    Converts a Niddel Magnet v2 API alert into an approximate CEF version 0 representation.
    :param obj: file-like object in binary mode to write to
    :param alert: dict containing a Niddel Magnet v2 API
    :return: an str / bytes object containing a CEF event
    """
    obj.write(header(device_vendor='Niddel', device_product='Magnet', device_version='1.0',
                     signature_id='infected_outbound',
                     name='Potentially Infected or Compromised Endpoint',
                     severity=max(ceil(alert['confidence'] / 10), 0)).encode('UTF-8'))

    ext = {
        'cs1': organization,
        'cs1Label': 'organizationId',
        'cs2': alert['batchDate'],
        'cs2Label': 'batchDate',
        'start': timestamp(alert['logDate'] + 'T' + alert['aggFirst']),
        'end': timestamp(alert['logDate'] + 'T' + alert['aggLast']),
        'externalId': alert['id'],
        'cfp1': alert['confidence'],
        'cfp1Label': 'confidence',
        'cnt': alert['aggCount'],
        'shost': alert.get('netSrcIpRdomain', None),
        'src': alert.get('netSrcIp', None),
        'dst': alert.get('netDstIp', None),
        'dhost': alert.get('netDstDomain', None),
        'dpt': alert.get('netDstPort', None),
        'proto': alert.get('netL4proto', None),
        'app': alert.get('netL7proto', alert.get('netApp', None)),
        'suid': alert.get('netSrcUser', None),
        'deviceCustomDate1': timestamp(alert.get('createdAt', None)),
        'deviceCustomDate1Label': 'createdAt',
        'deviceCustomDate2': timestamp(alert.get('updatedAt', None)),
        'deviceCustomDate2Label': 'updatedAt',
        'deviceDirection': 1,
        'dtz': 'GMT'
    }

    if 'netBlocked' in alert:
        if alert['netBlocked']:
            ext['act'] = 'allow'
        else:
            ext['act'] = 'deny'

    if 'tags' in alert:
        ext['cs3'] = ','.join(sorted(alert['tags']))
        ext['cs3Label'] = 'tags'

    if 'netDeviceTypes' in alert:
        ext['cs4'] = ','.join(sorted(alert['netDeviceTypes']))
        ext['cs4Label'] = 'netDeviceTypes'

    if 'netSrcProcessId' in alert:
        ext['cs5'] = alert['netSrcProcessId']
        ext['cs5Label'] = 'netSrcProcessId'

    # merge header and extension
    obj.write(extension(ext).encode('UTF-8'))




import datetime

import six
import iso8601


class UTC(datetime.tzinfo):
    """tzinfo derived concrete class for UTC"""
    _offset = datetime.timedelta(0)
    _dst = datetime.timedelta(0)
    _name = "UTC"

    def utcoffset(self, dt):
        return self.__class__._offset

    def dst(self, dt):
        return self.__class__._dst

    def tzname(self, dt):
        return self.__class__._name


UTC = UTC()

UTC_EPOCH = datetime.datetime(1970, 1, 1, tzinfo=UTC)


def seconds_from_UTC_epoch(value):
    if isinstance(value, six.string_types):
        value = iso8601.parse_date(value)
    elif not isinstance(value, datetime.datetime):
        raise ValueError('timestamp expected')
    return (value - UTC_EPOCH).total_seconds()



import io

from setuptools import setup
from magnetsdk2 import __version__


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


setup(
    name='magnetsdk2',
    description='Python SDK to access the Niddel Magnet API v2',
    long_description=read('README.rst'),
    author='Niddel Corp.',
    author_email='contact@niddel.com',
    version=__version__,
    url='https://github.com/niddel/magnet-api2-sdk-python/',
    license='Apache Software License',
    install_requires=['requests>=2.12.5,<3', 'six>=1.10,<2', 'iso8601>=0.1.12,<1',
                      'rfc3987>=1.3.7,<2', 'boto3>=1.4,<2'],
    tests_require=['pytest>=3.2.5,<4'],
    setup_requires=['pytest-runner>=3,<4'],
    packages=['magnetsdk2'],
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: English',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries',
    ],
    entry_points={
        'console_scripts': ['niddel=magnetsdk2.cli:main']
    }
)



# -*- coding: utf-8 -*-
"""
A simple client that allows idiomatic access to the Niddel Magnet v2 REST API
(https://api.niddel.com/v2). Uses the wonderful requests (http://docs.python-requests.org/)
package to perform the requests.
"""
__version__ = '1.4.1'

# set up package logger
import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

# import Connection class into namespace, but handle failure during setup if
# dependencies are not present yet
try:
    from magnetsdk2.connection import Connection
except ImportError:
    pass
